@@ -10,7 +10,7 @@ class RootController(object):
 
     @index.when(method='POST')
     def index_post(self, q):
-        redirect('http://pecan.readthedocs.org/en/latest/search.html?q=%s' % q)
+        redirect('https://pecan.readthedocs.io/en/latest/search.html?q=%s' % q)
 
     @expose('error.html')
     def error(self, status):
