@@ -1,13 +1,14 @@
-from json import dumps
+import time
+from json import dumps, loads
 import warnings
+from unittest import mock
 
 from webtest import TestApp
 from six import b as b_
 from six import u as u_
 import webob
-import mock
 
-from pecan import Pecan, expose, abort
+from pecan import Pecan, expose, abort, Request, Response
 from pecan.rest import RestController
 from pecan.hooks import PecanHook, HookController
 from pecan.tests import PecanTestCase
@@ -24,6 +25,21 @@ class TestThreadingLocalUsage(PecanTestCase):
                 assert isinstance(resp, webob.Response)
                 return 'Hello, World!'
 
+            @expose()
+            def warning(self):
+                return ("This should be unroutable because (req, resp) are not"
+                        " arguments.  It should raise a TypeError.")
+
+            @expose(generic=True)
+            def generic(self):
+                return ("This should be unroutable because (req, resp) are not"
+                        " arguments.  It should raise a TypeError.")
+
+            @generic.when(method='PUT')
+            def generic_put(self, _id):
+                return ("This should be unroutable because (req, resp) are not"
+                        " arguments.  It should raise a TypeError.")
+
         return RootController
 
     def test_locals_are_not_used(self):
@@ -36,6 +52,36 @@ class TestThreadingLocalUsage(PecanTestCase):
 
             self.assertRaises(AssertionError, Pecan, self.root)
 
+    def test_threadlocal_argument_warning(self):
+        with mock.patch('threading.local', side_effect=AssertionError()):
+
+            app = TestApp(Pecan(self.root(), use_context_locals=False))
+            self.assertRaises(
+                TypeError,
+                app.get,
+                '/warning/'
+            )
+
+    def test_threadlocal_argument_warning_on_generic(self):
+        with mock.patch('threading.local', side_effect=AssertionError()):
+
+            app = TestApp(Pecan(self.root(), use_context_locals=False))
+            self.assertRaises(
+                TypeError,
+                app.get,
+                '/generic/'
+            )
+
+    def test_threadlocal_argument_warning_on_generic_delegate(self):
+        with mock.patch('threading.local', side_effect=AssertionError()):
+
+            app = TestApp(Pecan(self.root(), use_context_locals=False))
+            self.assertRaises(
+                TypeError,
+                app.put,
+                '/generic/'
+            )
+
 
 class TestIndexRouting(PecanTestCase):
 
@@ -315,9 +361,13 @@ class TestControllerArguments(PecanTestCase):
         except Exception as ex:
             assert type(ex) == TypeError
             assert ex.args[0] in (
-                "index() takes exactly 4 arguments (3 given)",
-                "index() missing 1 required positional argument: 'id'"
-            )  # this messaging changed in Python 3.3
+                "index() takes exactly 2 arguments (1 given)",
+                "index() missing 1 required positional argument: 'id'",
+                (
+                    "TestControllerArguments.app_.<locals>.RootController."
+                    "index() missing 1 required positional argument: 'id'"
+                ),
+            )  # this messaging changed in Python 3.3 and again in Python 3.10
 
     def test_single_argument(self):
         r = self.app_.get('/1')
@@ -717,9 +767,13 @@ class TestControllerArguments(PecanTestCase):
         except Exception as ex:
             assert type(ex) == TypeError
             assert ex.args[0] in (
-                "eater() takes at least 4 arguments (3 given)",
-                "eater() missing 1 required positional argument: 'id'"
-            )  # this messaging changed in Python 3.3
+                "eater() takes exactly 2 arguments (1 given)",
+                "eater() missing 1 required positional argument: 'id'",
+                (
+                    "TestControllerArguments.app_.<locals>.RootController."
+                    "eater() missing 1 required positional argument: 'id'"
+                ),
+            )  # this messaging changed in Python 3.3 and again in Python 3.10
 
     def test_one_remainder(self):
         r = self.app_.get('/eater/1')
@@ -956,8 +1010,8 @@ class TestRestController(PecanTestCase):
             assert r.status_int == 302
 
     def test_invalid_custom_action(self):
-        r = self.app_.get('/things?_method=BAD', status=404)
-        assert r.status_int == 404
+        r = self.app_.get('/things?_method=BAD', status=405)
+        assert r.status_int == 405
 
     def test_named_action(self):
         # test custom "GET" request "length"
@@ -1310,3 +1364,85 @@ class TestHooks(PecanTestCase):
         assert run_hook[3] == 'inside_sub'
         assert run_hook[4] == 'after1'
         assert run_hook[5] == 'after2'
+
+
+class TestGeneric(PecanTestCase):
+
+    @property
+    def root(self):
+        class RootController(object):
+
+            def __init__(self, unique):
+                self.unique = unique
+
+            @expose(generic=True, template='json')
+            def index(self, req, resp):
+                assert self.__class__.__name__ == 'RootController'
+                assert isinstance(req, Request)
+                assert isinstance(resp, Response)
+                assert self.unique == req.headers.get('X-Unique')
+                return {'hello': 'world'}
+
+            @index.when(method='POST', template='json')
+            def index_post(self, req, resp):
+                assert self.__class__.__name__ == 'RootController'
+                assert isinstance(req, Request)
+                assert isinstance(resp, Response)
+                assert self.unique == req.headers.get('X-Unique')
+                return req.json
+
+            @expose(template='json')
+            def echo(self, req, resp):
+                assert self.__class__.__name__ == 'RootController'
+                assert isinstance(req, Request)
+                assert isinstance(resp, Response)
+                assert self.unique == req.headers.get('X-Unique')
+                return req.json
+
+            @expose(template='json')
+            def extra(self, req, resp, first, second):
+                assert self.__class__.__name__ == 'RootController'
+                assert isinstance(req, Request)
+                assert isinstance(resp, Response)
+                assert self.unique == req.headers.get('X-Unique')
+                return {'first': first, 'second': second}
+
+        return RootController
+
+    def test_generics_with_im_self_default(self):
+        uniq = str(time.time())
+        with mock.patch('threading.local', side_effect=AssertionError()):
+            app = TestApp(Pecan(self.root(uniq), use_context_locals=False))
+            r = app.get('/', headers={'X-Unique': uniq})
+            assert r.status_int == 200
+            json_resp = loads(r.body.decode())
+            assert json_resp['hello'] == 'world'
+
+    def test_generics_with_im_self_with_method(self):
+        uniq = str(time.time())
+        with mock.patch('threading.local', side_effect=AssertionError()):
+            app = TestApp(Pecan(self.root(uniq), use_context_locals=False))
+            r = app.post_json('/', {'foo': 'bar'}, headers={'X-Unique': uniq})
+            assert r.status_int == 200
+            json_resp = loads(r.body.decode())
+            assert json_resp['foo'] == 'bar'
+
+    def test_generics_with_im_self_with_path(self):
+        uniq = str(time.time())
+        with mock.patch('threading.local', side_effect=AssertionError()):
+            app = TestApp(Pecan(self.root(uniq), use_context_locals=False))
+            r = app.post_json('/echo/', {'foo': 'bar'},
+                              headers={'X-Unique': uniq})
+            assert r.status_int == 200
+            json_resp = loads(r.body.decode())
+            assert json_resp['foo'] == 'bar'
+
+    def test_generics_with_im_self_with_extra_args(self):
+        uniq = str(time.time())
+        with mock.patch('threading.local', side_effect=AssertionError()):
+            app = TestApp(Pecan(self.root(uniq), use_context_locals=False))
+            r = app.get('/extra/123/456', headers={'X-Unique': uniq})
+            assert r.status_int == 200
+            json_resp = loads(r.body.decode())
+            assert json_resp['first'] == '123'
+            assert json_resp['second'] == '456'
